import os
import shutil
import zipfile
import logging
import boto3
from boto3.s3.transfer import (
    S3Transfer,
    TransferConfig,
)
from botocore.exceptions import BotoCoreError

logger = logging.getLogger(__name__)


class BaseEngine:

    NAME = None

    def __init__(self, directory):
        self.directory = directory
        self.markets_loaded = []
        self.validate_settings()

    def __call__(self, market_id, market_definition, stream_id):
        logger.info('Loading %s to %s:%s' % (market_id, self.NAME, self.directory))
        file_dir = os.path.join('/tmp', stream_id, market_id)

        # check that market has not already been loaded
        if market_id in self.markets_loaded:
            logger.error('File: /tmp/%s/%s has already been loaded' % (stream_id, market_id))
            return

        # check that file actually exists
        if not os.path.isfile(file_dir):
            logger.error('File: %s does not exist in /tmp/%s/' % (market_id, stream_id))
            return

        # zip file
        zip_file_dir = self.zip_file(file_dir, market_id, stream_id)
        # core load code
        self.load(zip_file_dir, market_definition)
        # clean up
        self.clean_up(file_dir, zip_file_dir)

        self.markets_loaded.append(market_id)

    def load(self, zip_file_dir, market_definition):
        """Loads zip_file_dir to expected dir
        """
        raise NotImplementedError

    def validate_settings(self):
        """Validates settings e.g. directory
        provided, raises OSError
        """
        pass

    @staticmethod
    def create_metadata(market_definition):
        try:
            del market_definition['runners']
        except KeyError:
            pass
        return dict([a, str(x)] for a, x in market_definition.items())

    @staticmethod
    def clean_up(file_dir, zip_file_dir):
        """remove txt file and zip from /tmp
        """
        logger.info('Removing txt and zip')
        os.remove(file_dir)
        os.remove(zip_file_dir)

    @staticmethod
    def zip_file(file_dir, market_id, stream_id):
        """zips txt file into filename.zip
        """
        zip_file_directory = os.path.join('/tmp', stream_id, '%s.zip' % market_id)
        zf = zipfile.ZipFile(zip_file_directory, mode='w')
        zf.write(file_dir, os.path.basename(file_dir), compress_type=zipfile.ZIP_DEFLATED)
        zf.close()
        return zip_file_directory

    @property
    def markets_loaded_count(self):
        return len(self.markets_loaded)

    @property
    def extra(self):
        return {
            'name': self.NAME,
            'markets_loaded': self.markets_loaded,
            'markets_loaded_count': self.markets_loaded_count,
            'directory': self.directory,
        }


class Local(BaseEngine):

    NAME = 'LOCAL'

    def validate_settings(self):
        if not os.path.isdir(self.directory):
            raise OSError('File dir %s does not exist' % self.directory)

    def load(self, zip_file_dir, market_definition):
        shutil.copy(zip_file_dir, self.directory)


class S3(BaseEngine):

    NAME = 'S3'

    def __init__(self, directory, access_key=None, secret_key=None):
        self.s3 = boto3.client(
            's3',
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
        )
        transfer_config = TransferConfig(use_threads=False)
        self.transfer = S3Transfer(self.s3, config=transfer_config)
        super(S3, self).__init__(directory)

    def validate_settings(self):
        # error raised if bucket not present
        self.s3.head_bucket(Bucket=self.directory)

    def load(self, zip_file_dir, market_definition):
        event_type_id = market_definition.get('eventTypeId')
        try:
            self.transfer.upload_file(
                filename=zip_file_dir,
                bucket=self.directory,
                key=os.path.join(
                    'marketdata', 'streaming', event_type_id, os.path.basename(zip_file_dir)
                ),
                extra_args={
                    'Metadata': self.create_metadata(market_definition)
                }
            )
            logger.info('%s successfully loaded to s3' % zip_file_dir)
        except (BotoCoreError, Exception) as e:
            logger.error('Error loading to s3: %s' % e)



import os
import logging
import threading
from betfairlightweight import (
    APIClient,
    BetfairError,
)

from .listener import FlumineListener
from .exceptions import RunError

logger = logging.getLogger(__name__)


class Flumine:

    def __init__(self, recorder, settings=None, unique_id=1e3):
        self.trading = self._create_client(settings)
        self.recorder = recorder
        self.unique_id = unique_id

        self._running = False
        self._socket = None
        self.listener = FlumineListener(recorder)  # output queue hack

    def start(self, heartbeat_ms=None, conflate_ms=None, segmentation_enabled=None, async=True):
        """Checks trading is logged in, creates socket,
        subscribes to markets, sets running to True and
        starts handler/run threads.
        """
        logger.info('Starting stream: %s' % self.unique_id)
        if self._running:
            raise RunError('Flumine is already running, call .stop() first')
        self._check_login()
        self._create_socket()
        self.unique_id = self._socket.subscribe_to_markets(
                market_filter=self.recorder.market_filter,
                market_data_filter=self.recorder.market_data_filter,
                conflate_ms=conflate_ms,
                heartbeat_ms=heartbeat_ms,
                segmentation_enabled=segmentation_enabled,
        )
        self._running = True
        if async:
            threading.Thread(target=self._run, daemon=True).start()
        else:
            self._run()

    def stop(self):
        """Stops socket, sets running to false
        and socket to None
        """
        logger.info('Stopping stream: %s' % self.unique_id)
        if not self._running:
            raise RunError('Flumine is not running')
        if self._socket:
            self._socket.stop()
        self._running = False
        self._socket = None

    def stream_status(self):
        """Checks sockets status
        """
        return str(self._socket) if self._socket else 'Socket not created'

    @staticmethod
    def _create_client(settings):
        """Returns APIClient based on settings
        or looks for username in os.environ
        """
        if settings:
            return APIClient(
                **settings.get('betfairlightweight')
            )
        else:
            username = os.environ.get('username')
            return APIClient(
                username=username
            )

    def _run(self):
        """ Runs socket and catches any errors
        """
        try:
            self._socket.start(async=False)
        except BetfairError as e:
            logger.error('Betfair error: %s' % e)
            self.stop()
        except Exception as e:
            logger.error('Unknown error: %s' % e)
            self.stop()

    def _check_login(self):
        """Login if session expired
        """
        if self.trading.session_expired:
            self.trading.login()

    def _create_socket(self):
        """Creates stream
        """
        self._socket = self.trading.streaming.create_stream(
            unique_id=self.unique_id,
            description='Flumine Socket',
            listener=self.listener
        )

    @property
    def running(self):
        return True if self._running else False

    @property
    def status(self):
        return 'running' if self._running else 'not running'

    def __repr__(self):
        return '<Flumine>'

    def __str__(self):
        return '<Flumine [%s]>' % self.status



import logging

from .flumine import Flumine
from . import resources
from .exceptions import FlumineException


__title__ = 'flumine'
__version__ = '0.2.0'
__author__ = 'Liam Pauling'


# Set default logging handler to avoid "No handler found" warnings.
class NullHandler(logging.Handler):
    def emit(self, record):
        pass

logging.getLogger(__name__).addHandler(NullHandler())



from betfairlightweight.streaming.listener import StreamListener
from betfairlightweight.streaming.stream import BaseStream


"""
Custom listener that doesn't do any processing,
helps reduce CPU. Hacks the output queue with a
recorder to remove the need for a handler thread.
"""


class FlumineListener(StreamListener):

    def __init__(self, recorder, max_latency=0.5):
        super(FlumineListener, self).__init__(max_latency)
        self.recorder = recorder

    def _add_stream(self, unique_id, stream_type):
        if stream_type == 'marketSubscription':
            return FlumineStream(
                unique_id, self.recorder, self.max_latency, self.lightweight
            )
        elif stream_type == 'orderSubscription':
            raise ValueError('Not expecting an order stream...')


class FlumineStream(BaseStream):

    def _process(self, market_books, publish_time):
        for market_book in market_books:
            market_id = market_book.get('id')
            if 'marketDefinition' in market_book and market_book['marketDefinition']['status'] == 'CLOSED':
                if market_id in self._caches:
                    # removes closed market from cache
                    del self._caches[market_id]
            elif self._caches.get(market_id) is None:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()

        self.output_queue(market_books, publish_time)
        self._updates_processed += len(market_books)

    def __str__(self):
        return 'FlumineStream'

    def __repr__(self):
        return '<FlumineStream [%s]>' % len(self._caches)



import os
import json
import logging
from betfairlightweight.filters import (
    streaming_market_filter,
    streaming_market_data_filter,
)

from ..utils import create_short_uuid

logger = logging.getLogger(__name__)


class BaseRecorder:
    """Base recorder which connects to all
    markets by default.
    """

    NAME = 'BASE_RECORDER'

    def __init__(self, storage_engine, market_filter=None, market_data_filter=None):
        self.storage_engine = storage_engine
        self.market_filter = market_filter or streaming_market_filter()
        self.market_data_filter = market_data_filter or streaming_market_data_filter(
            fields=[
                'EX_ALL_OFFERS', 'EX_TRADED', 'EX_TRADED_VOL', 'EX_LTP', 'EX_MARKET_DEF', 'SP_TRADED', 'SP_PROJECTED'
            ]
        )
        self.stream_id = create_short_uuid()
        self._setup()
        logger.info('Recorder created %s' % self.stream_id)

    def __call__(self, market_book, publish_time):
        """Checks market using market book parameters
        function then passes market_book to be processed.

        :param market_book: Market Book object
        :param publish_time: Publish time of market book
        """
        if self.market_book_parameters(market_book):
            self.process_market_book(market_book, publish_time)

    def market_book_parameters(self, market_book):
        """Logic used to decide if market_book should
        be processed

        :param market_book: Market Book object
        :return: True if market is to be processed
        """
        return True

    def process_market_book(self, market_book, publish_time):
        """Function that processes market book

        :param market_book: Market Book object
        :param publish_time: Publish time of market book
        """
        raise NotImplementedError

    def on_market_closed(self, market_book):
        """Function run when market is closed.
        """
        market_id = market_book.get('id')
        market_definition = market_book.get('marketDefinition')
        logger.info('Closing market %s' % market_id)
        self.storage_engine(market_id, market_definition, self.stream_id)

    def _setup(self):
        """Create stream folder in /tmp  # todo
        """
        directory = os.path.join('/tmp', self.stream_id)
        if not os.path.exists(directory):
            os.makedirs(directory)

    def __str__(self):
        return '<%s>' % self.NAME


class StreamRecorder(BaseRecorder):
    """Data recorder, records stream data
    to /tmp/market_id
    """

    NAME = 'STREAM_RECORDER'

    def process_market_book(self, market_book, publish_time):
        for market in market_book:
            filename = '%s' % market.get('id')
            file_directory = os.path.join('/tmp', self.stream_id, filename)

            with open(file_directory, 'a') as outfile:
                outfile.write(
                    json.dumps({
                        "op": "mcm",
                        "clk": None,
                        "pt": publish_time,
                        "mc": [market]
                    }) + '\n'
                )

            if 'marketDefinition' in market and market['marketDefinition']['status'] == 'CLOSED':
                self.on_market_closed(market)





class FlumineException(Exception):
    """Base class for Flumine Errors"""
    pass


class RunError(FlumineException):
    """Error raised if start() called when already running"""

    def __init__(self, message):
        super(RunError, self).__init__(message)



import uuid
from betfairlightweight.filters import streaming_market_filter


def create_short_uuid():
    return str(uuid.uuid4())[:8]


def gb_ie_racing():
    return streaming_market_filter(
        event_type_ids=['7'],
        country_codes=['GB', 'IE'],
        market_types=['WIN'],
    )


def au_racing():
    return streaming_market_filter(
        event_type_ids=['7'],
        country_codes=['AU'],
        market_types=['WIN'],
    )


def us_racing():
    return streaming_market_filter(
        event_type_ids=['7'],
        country_codes=['US'],
        market_types=['WIN'],
    )



import re
from setuptools import setup


INSTALL_REQUIRES = [
    'betfairlightweight>=1.2',
    'boto3',
]

TEST_REQUIRES = [
]

with open('flumine/__init__.py', 'r') as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        f.read(),
        re.MULTILINE
    ).group(1)

setup(
        name='flumine',
        version=version,
        packages=[
                'flumine',
                'flumine.resources',
                'flumine.storage'
        ],
        package_dir={
                'flumine': 'flumine'
        },
        install_requires=INSTALL_REQUIRES,
        url='https://github.com/liampauling/flumine',
        license='MIT',
        author='liampauling',
        author_email='',
        description='Betfair data record framework utilising betfair streaming and flask',
        classifiers=[
                'License :: OSI Approved :: MIT License',
                'Operating System :: OS Independent',
                'Programming Language :: Python :: 3.4',
                'Programming Language :: Python :: 3.5',
                'Programming Language :: Python :: 3.6',
        ],
        test_suite='tests'
)
